/******************************************************************************
 * Lemonade JS
 * https://lemonadejs.com
 *
 * ----
 *
 * Mysql link loader script
 *
 *****************************************************************************/

function mysqlLoader () {}

/**
 * Master config
 * @type {{}}
 */
mysqlLoader.prototype.master = {};

/**
 * Slave config
 * @type {{}}
 */
mysqlLoader.prototype.slave = {};

/**
 * Default db
 * @type {string}
 */
mysqlLoader.prototype.defaultdb = '';

/**
 * References to adaptors (a new adaptor is
 * created for each unique db)
 * @type {{}}
 */
mysqlLoader.prototype.adaptors = {};

/**
 * Try to execute a simple query on both master and slave
 * @param {object} master
 * @param {object} slave
 * @param {string} db
 * @param {function} cb
 */
mysqlLoader.prototype.load = function(master, slave, db, cb) {
    if (!master && !slave) {
        throw new Error(
            'Mysql invalid config. Define host/user/pass/port ' +
            'in the configuration file'
        );
    }
    if (master) {
        this.master = master;
    }
    if (slave) {
        this.slave = slave;
    }
    if (!db) {
        throw new Error(
            'Mysql invalid config. Missing default db'
        );
    }
    this.defaultdb = db;
    /**
     * Prepare a cb to be ran after executing a query on the master
     * and one on the slave
     */
    var notoload = 2, retmsg = '', precb = function(err, msg) {
        if (err) {
            cb(err);
        } else {
            if (msg) {
                retmsg += msg + ',';
            }
            if (!--notoload) {
                cb(null, '(' + retmsg.substring(0, retmsg.length - 1) + ') pool up');
            }
        }
    };
    /**
     * Try the query on the master
     */
    this.link(db).query(
        '/*MASTER*/ SELECT 1'
        ,[]
        ,function(err) {
            precb(err, 'master');
        }
    );
    /**
     * If we got a non-replication setup don't bother
     */
    if (slave && master !== slave) {
        this.link(db).query(
            'SELECT 1'
            ,[]
            ,function(err) {
                precb(err, 'slave');
            }
        );
    } else {
        precb(null, '');
    }
};

/**
 * Return a singleton for the adaptor based on database (there will be a
 * different connection pool and adaptor object for each db name used)
 * @param {string} db
 * @return {object}
 */
mysqlLoader.prototype.link = function(db) {
    db = db || this.defaultdb;
    if (!this.adaptors[db]) {
        this.adaptors[db] = this.kernel.include(this.kernel.kerneldir + 'adaptors/mysql').singleton(db,
            this.master, this.slave, db
        );
    }
    return this.adaptors[db];
}

/**
 * Gather up statuses from all open adaptors
 * @return {object}
 */
mysqlLoader.prototype.status = function() {
    var merge = {};
    for (var db in this.adaptors) {
        merge[db] = this.adaptors[db].status;
    }
    return merge;
}

/**
 * Unload all adaptors (prepare for shutdown)
 * @param {function} cb
 */
mysqlLoader.prototype.unload = function(cb) {
    noloaded = Object.keys(this.adaptors).length;
    if (!noloaded) {
        cb(null, 'pool not created');
    } else {
        for (var db in this.adaptors) {
            this.adaptors[db].shutdown(function() {
                if (!--noloaded) {
                    cb(null, 'pool drained');
                }
            }.bind(this));
        }
    }
};

/**
 * Export
 */
exports = module.exports = mysqlLoader;