/*****************************************************************************
 * Lemonade JS
 * https://lemonadejs.com
 *
 * ----
 *
 * Rabbit
 *
 *  Adaptor to connect to Rabbitmq. Subscribe to a queue and have messages
 *  pushed. Publish messages.
 *
 * @param {Object} host
 *
 *****************************************************************************/

function Rabbit (host) {
    if (typeof host !== 'object') {
        host = { host : host };
    }
    this.host = host;
    this.host.defaultExchangeName = ""; //bind directly to queues
    this.connect();
}

/**
 * Module dependencies
 */
var amqp = require('amqp');

/**
 * Connection events
 * @type {string}
 */
Rabbit.prototype.CONNECTED    = 'rabbitmq-connected';
Rabbit.prototype.DISCONNECTED = 'rabbitmq-disconnected';
Rabbit.prototype.ERROR        = 'rabbitmq-error';

/**
 * Host to connect to
 * @type {object}
 */
Rabbit.prototype.host = {
      host      : 'localhost'
    , port      : 5672
    , login     : 'guest'
    , password  : 'guest'
    , vhost     : '/'
};

/**
 * Amqp connection object
 * @type {{}}
 * @private
 */
Rabbit.prototype._connection;

/**
 * Prepared queues
 * @type {{}}
 * @private
 */
Rabbit.prototype._prepared = {};

/**
 * Subscribed queues
 * @type {{}}
 * @private
 */
Rabbit.prototype._subscribed = {};

/**
 * Prefetch count
 * @type {number}
 * @private
 */
Rabbit.prototype._prefetch = 10;

/**
 * Queue init params (todo : make this configurable)
 * @type {{autoDelete: boolean, durable: boolean}}
 * @private
 */
Rabbit.prototype._qparams = { autoDelete : false, durable : true };

/**
 * Status object
 * @returns {{}}
 */
Rabbit.prototype.status = function() {
    return {
         subscribed  : this._subscribed
        ,published   : this._prepared
    }
};

/**
 * Create the connection and listen for connection ready
 * and connection closed
 */
Rabbit.prototype.connect = function() {

    this._connection = amqp.createConnection(this.host, { reconnect : false });

    /**
     * Wait for ready signal. Emit custom event
     * when ready.
     */
    this._connection.on('ready', function(msg) {
        this.kernel.emit(this.CONNECTED, msg);
    }.bind(this));

    /**
     * Wait for connection closed
     */
    this._connection.on('closed', function(msg) {
        this.kernel.emit(this.DISCONNECTED, msg);
    }.bind(this));

    /**
     * Error listener
     */
    this._connection.on('error', function(err) {
        this.kernel.emit(this.ERROR, err);
    }.bind(this));

};

/**
 * Gracefully end the connection
 */
Rabbit.prototype.shutdown = function() {
    for (var key in this._subscribed) {
        this.unsubscribe(key, function() {});
    }
    this._connection.end();
    this.kernel.emit(this.DISCONNECTED);
};

/**
 * Subscribe to a queue. Will receive messages as they are published. Instant ack.
 *
 * @param {string} queue
 * @param {function} callbackfetch
 * @param {function} callbacksubscribe
 * @return
 */
Rabbit.prototype.subscribe = function(queue, callbackfetch, callbacksubscribe) {
    this._connection.queue(queue, this._qparams , function(q) {

        /**
         * Mark queue as prepared (needed for publishing)
         */
        this._prepared[queue] = {
             queue     : queue
            ,published : 0
        };

        q.subscribe({ ack : true, prefetchCount : this._prefetch }, function (message, headers, deliveryInfo) {

            /**
             * Increment no. of fetched messages for this queue
             */
            if (typeof this._subscribed[queue] !== 'undefined') {
                this._subscribed[queue].fetched++;
            }

            /**
             * If the message retrieved is and object, transform it to string
             */
            callbackfetch(null, (typeof message.data !== 'undefined') ? message.data.toString() : message);

            /**
             * Move to the next message
             */
            q.shift();

        }.bind(this)).addCallback(function(ok) {

            /**
             * Keep the consumer tag for debugging
             */
            if (typeof this._subscribed[queue] === 'undefined') {
                this._subscribed[queue] = {
                     queue   : queue
                    ,tag     : ok.consumerTag
                    ,active  : true
                    ,fetched : 1
                    ,objq    : q
                };
            } else {
                this._subscribed[queue].active = true;
                this._subscribed[queue].tag    = ok.consumerTag;
            }

            callbacksubscribe(ok);

        }.bind(this));
    }.bind(this));
};

/**
 * Unsubscribe from a queue
 *
 * @param {string} queue
 * @param {function} callback
 * @return
 */
Rabbit.prototype.unsubscribe = function(queue, callback) {
    if (typeof this._subscribed[queue] !== 'undefined') {
        this._subscribed[queue].objq.unsubscribe(this._subscribed[queue].tag);
        this._subscribed[queue].active = false;
        callback();
    }
};

/**
 * Publish message to queue
 *
 * @param {string} queue
 * @param {string} message
 * @param {function} callback
 * @return
 */
Rabbit.prototype.publish = function(queue, message, callback) {
    if (typeof this._prepared[queue] === 'undefined') {
        /**
         * Initiate a connection to the queue first to create it if it's not
         * already there
         */
        this._connection.queue(queue, {autoDelete : false, durable : true}, function() {

            this._connection.publish(queue, message);
            this._prepared[queue] = {
                 queue     : queue
                ,published : 1
            };
            callback();

        }.bind(this));
    } else {

        this._connection.publish(queue, message);
        this._prepared[queue].published++;
        callback();

    }
};

/**
 * Export
 */
exports = module.exports = Rabbit;